/*
   HP Occlusion Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to perform Occlusion testing in OpenGL using the
   HP_occlusion_test extension.  This is simlar to the last occlusion demo but
   using the HP_occlusion_test extension we can't tell how many pixels are viewable
   just if it is visible or not.  This is only 1 of the limitations of this occlusion
   test.  Another downer is that in order to tell if an object is visible or not correctly
   we must draw the scene in front/back order.  This means that everthing close to the
   camera must be drawn first and in order.  Now this is not as big a deal if you are rendering
   a level using a bsp tree but for everything else its a big problem because determining
   which object should be drawn first and which afterwards will take too long.  Long enough
   to defeat the purpose of this.  Another bad thing about HP_occlusion_test is that you can
   only test one object at a time.  In the last occlusion demo we tested them all at the
   same time.  For this demo I set it up to render the biggest object first then the
   smaller ones.  This way if a smaller object is in front of a bigger one you can still see
   the bigger object.  If the bigger object is rendered first then the smaller one will be
   rejected.  The bad thing is that the more objects you have the harder it is to keep up
   with this.  With that known this occlusion technique is not always the best.
   If you have no other choice then this technique could help some but the NV_occlusion_query
   will win the fight every time.
*/


// Demo: HP Occlusion.


#define WIN32_LEAN_AND_MEAN// Trims down the libraries.
#define VC_LEANMEAN        // Trims even further.


#include<windows.h>        // Windows header file.
#include<stdio.h>          // Standard input/output.
#include<gl/gl.h>          // OpenGL header.
#include<gl/glu.h>         // OpenGL utility.
#include<gl/glaux.h>       // OpenGL auxiluary.
#include<gl/glext.h>       // OpenGL extensions.
#include"CVector.h"        // Vector class.
#include"CTGALoader.h"     // Texture object CTexture.
#include"CGLFont.h"        // OpenGL font class.


// As I mentioned with the HP Occlusion testing that we will have to render
// the bounding geometry into the test.  Since all of our objects are simple
// this is not a big deal now but if we had complex models or level geometry
// then using the bounding box or bounding sphere only makes sense for
// performance reasons instead of drawing the whole model.
struct BoundingBox
{
   CVector4 min;           // min x, y, z values of the object.
   CVector4 max;           // max x, y, z values of the object.
};

// I created this enum to hold the array index for each of the objects we are
// rendering in the scene.  This makes things easier for me.
enum RenderingObjects
{
   SQUARE = 0,
   SPHERE,
   CUBE,
   TORUS,
};


// Function prototypes...
bool InitializeGL();                   // Initialize OpenGL.
void RenderScene();                    // Renders each frame.
void DrawGround();                     // Renders the ground surface.
void Render(int i);                    // Renders all objects.
bool RenderTest(BoundingBox &Box);     // Used to test if a object is visible.
void DrawBoundingBox(BoundingBox &Box);// Draw a bounding box.
void Shutdown();                       // Used to clear out everything we used.


HDC g_HDC;                             // Global HDC used in the RenderScene() to swap buffers.

#define MAX_POINTS 4                   // The max points in the square object.
#define MAX_OBJECTS 4                  // Max amount of objects in scene.

// The normal of the square object (the quad).
CVector4 SquareNormal(0.0f, 0.0f, 1.0f, 0.0f);

// This is a list of all the objects positions in the scene.
CVector4 ObjectPositions[MAX_OBJECTS];

// This stores a list of each objects' size in the scene.
float ObjectSizes[MAX_OBJECTS];

// Font object used to display text to the screen.
CGLFont font;

// Used to store a string to print output with.
char output[50] = {0};

// This is one of the objects used in this demo.  It is a simple square.
CVector4 Square[MAX_POINTS];
CTexCoord SquareCoords[MAX_POINTS];

// This is a list of bounding box values for each object in the scene.
BoundingBox ObjectsBBox[MAX_OBJECTS];

// This simple stores a "true" or "false" string for each object for whether it is visible or not.
char Visible[MAX_OBJECTS][10] = {0};

// xRotation and yRotation is used to rotate the entire scene.
float xRotation = 0.0f;
float yRotation = 0.0f;

CTGALoader Terrain;          // Terrain texture image.
CTGALoader SquareTex;        // Texture for the square object.

// Used to test if the hardware support the occlusion extension.
bool OcclusionSupported = false;

// Used to Render the bounding boxes to the screen if true.  Output purposes only.
bool RenderBoundingBoxes = false;


void SetupPixelFormat(HDC hDC)
{
   int nPixelFormat;

   static PIXELFORMATDESCRIPTOR pfd = {
         sizeof(PIXELFORMATDESCRIPTOR),   // size of structure.
         1,                               // always 1.
         PFD_DRAW_TO_WINDOW |             // support window
         PFD_SUPPORT_OPENGL |             // support OpenGl
         PFD_DOUBLEBUFFER,                // support double buffering
         PFD_TYPE_RGBA,                   // support RGBA
         16,                              // bit color mode
         0, 0, 0, 0, 0, 0,                // ignore color bits
         0,                               // no alpha buffer
         0,                               // ignore shift bit
         0,                               // no accumulation buffer
         0, 0, 0, 0,                      // ignore accumulation bits.
         16,                              // number of depth buffer bits.
         0,                               // number of stencil buffer bits.
         0,                               // 0 means no auxiliary buffer
         PFD_MAIN_PLANE,                  // The main drawing plane
         0,                               // this is reserved
         0, 0, 0 };                       // layer masks ignored.

   // this chooses the best pixel format and returns index.
   nPixelFormat = ChoosePixelFormat(hDC, &pfd);

   // This set pixel format to device context.
   SetPixelFormat(hDC, nPixelFormat, &pfd);
}


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   static HGLRC hRC;             // Rendering context.
   static HDC hDC;               // Device context.
   int width, height;            // The window width and height.
	static POINT oldMousePos;     // Last mouse position.
   static POINT currentMousePos; // Current mouse position.
   static bool  isMouseActive;   // Is the left mouse button down.


   switch(message)
      {
         case WM_CREATE:                     // Windows creation.
            hDC = GetDC(hwnd);               // This gets the device context for our window.
            g_HDC = hDC;                     // Assigns the global device context to this one.
            SetupPixelFormat(hDC);           // Call the pixel format function.

            hRC = wglCreateContext(hDC);     // Creates the rendering context.
            wglMakeCurrent(hDC, hRC);        // Makes the rendering context.
            return 0;
            break;


         case WM_CLOSE:                      // Close message.
         case WM_DESTROY:
            wglMakeCurrent(hDC, NULL);
            wglDeleteContext(hRC);           // Deletes the rendering context.

            PostQuitMessage(0);              // Says close the program.
            return 0;
            break;


         case WM_SIZE:                       // re-size message.
            height = HIWORD(lParam);         // This gets the height of the window.
            width = LOWORD(lParam);          // This gets the width of the window.

            if(height==0)                    // we don't want it to be possible for a
               {                             // height of 0.  If it is 0 me make it 1.
                  height = 1;
               }

            glViewport(0, 0, width, height); // resets the viewport to new dimensions.
            glMatrixMode(GL_PROJECTION);     // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            // calculate the aspect ratio of the window.
            gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, 1000.0f);

            glMatrixMode(GL_MODELVIEW);      // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            font.SetWidthHeight(width,
                                height);     // Send width and height so font is correctly drawn.

            return 0;
            break;


         case WM_KEYDOWN:
		      switch(wParam)
               { 
                  // Close the app if the user presses escape.
		            case VK_ESCAPE:
                     PostQuitMessage(0);
				         break;

                  case VK_SPACE:
                     RenderBoundingBoxes = !RenderBoundingBoxes;
                     break;

                  default:
                     break;
               }
            break;


			case WM_LBUTTONDOWN:
				oldMousePos.x = currentMousePos.x = LOWORD (lParam);
				oldMousePos.y = currentMousePos.y = HIWORD (lParam);
            isMouseActive = true;
			   break;


			case WM_LBUTTONUP:
				isMouseActive = false;
				break;


			case WM_MOUSEMOVE:
				currentMousePos.x = LOWORD (lParam);
				currentMousePos.y = HIWORD (lParam);

				if(isMouseActive)
					{
						xRotation -= (currentMousePos.x - oldMousePos.x);
						yRotation -= (currentMousePos.y - oldMousePos.y);
					}

				oldMousePos.x = currentMousePos.x;
				oldMousePos.y = currentMousePos.y;
			   break;


         default: // Always have a default just in case.
            break;
      }

   // What this does is pass all of the unhandled messages to DefWindowProc.
   return (DefWindowProc(hwnd, message, wParam, lParam));
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                // A message variable.
   WNDCLASSEX windowClass; // Your Window class.
   HWND hwnd;              // The Window handle.
   bool isFinished;        // This will be used to check if the program is done or not.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);        // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;    // style of the window.
   windowClass.lpfnWndProc = WndProc;              // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                     // Extra class information.
   windowClass.cbWndExtra = 0;                     // Extra window information.
   windowClass.hInstance = hInstance;              // Handle of application Instance.
   // Handle of application Icon.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW);// mouse cursor
   windowClass.hbrBackground = NULL;               // background color.
   windowClass.lpszMenuName = NULL;                // name of the main menu.
   windowClass.lpszClassName = "UltimateGameProgrammingClass";// window class name.
   // icon when minimized.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass))
      return 0;

   // Create window...
   hwnd = CreateWindowEx(NULL,// The extended window style.
                        "UltimateGameProgrammingClass",// window Class name.
                        "HP Occlusion Demo - By the Programming Ace.",// window name.
                        WS_OVERLAPPEDWINDOW | WS_VISIBLE |// The window style.
                        WS_SYSMENU | WS_CLIPCHILDREN |// window style.
                        WS_CLIPSIBLINGS,// window style.
                        100, 100,// window x, y coordinate.
                        640,
                        480,// window width and height.
                        NULL,// handle to parent window.
                        NULL,// menu.
                        hInstance,// handle to app instance.
                        NULL);   // pointer to window creation data.
   
   // If there was an error with creating the window, then close the program.
   if(!hwnd)
      return 0;

   ShowWindow(hwnd, SW_SHOW); // This shows the window.
   UpdateWindow(hwnd);        // This forces a paint message.

   isFinished = false;        // False = running, True = not running.

   // If initialize fail (return false), then we don't want the program to run.
   if(!InitializeGL())
      isFinished = true;

   // This is the messsage loop.
   while(!isFinished)
      {
         if(PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     isFinished = true;
                  }

               TranslateMessage(&msg);  // Translate any messages.
               DispatchMessage(&msg);   // Dispatch any messages.
            }
         else
            {
               RenderScene();          // Render a frame.
            }
      }

   // Shutdown (free) all resources we used up.
   Shutdown();

   return (int)msg.wParam;
}


bool InitializeGL()
{
   glClearColor(0.5f, 0.5f, 1.0f, 1.0f);  // Clear the screen to a set color.

   glShadeModel(GL_SMOOTH);               // Smooth shading in our scenes.
   glEnable(GL_LIGHTING);						//	Enable lighting.
	glEnable(GL_DEPTH_TEST);					// Enable depth testing.
	glEnable(GL_TEXTURE_2D);					// Enable texture mapping.

   // Setup the light source below.
   float ambientLight[] = {1.0f, 1.0f, 1.0f, 1.0f};
   float diffuseLight[] = {1.0, 1.0, 1.0, 1.0}; 
   float specularLight[] = {1.0, 1.0, 1.0, 1.0};

   glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuseLight);
   glLightfv(GL_LIGHT0, GL_SPECULAR, specularLight);
   glLightfv(GL_LIGHT0, GL_AMBIENT, ambientLight);
   glEnable(GL_LIGHT0);

   // Initialize the font object so we can display output using it.
   if(!font.Initialize(g_HDC, 20, "Arial")) return false;

   // Load the terrain texture.   
   if(!Terrain.LoadTGA("ground.tga")) return false;
   glGenTextures(1, &Terrain.ID);
   glBindTexture(GL_TEXTURE_2D, Terrain.ID);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

   int type = Terrain.type == 3 ? GL_RGB : GL_RGBA;
   gluBuild2DMipmaps(GL_TEXTURE_2D, type, Terrain.imageWidth, Terrain.imageHeight,
                     type, GL_UNSIGNED_BYTE, Terrain.image);

   // Load the square texture.
   if(!SquareTex.LoadTGA("decal.tga")) return false;
   glGenTextures(1, &SquareTex.ID);
   glBindTexture(GL_TEXTURE_2D, SquareTex.ID);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

   type = Terrain.type == 3 ? GL_RGB : GL_RGBA;
   gluBuild2DMipmaps(GL_TEXTURE_2D, type, SquareTex.imageWidth,
                     SquareTex.imageHeight,
                     type, GL_UNSIGNED_BYTE, SquareTex.image);

   // Initialize the square data.  This is taken out of the shadow volume demo.
	Square[0] = CVector4(1.0f, 2.0f, 1.0f);
   SquareCoords[0].tu = 1.0f; SquareCoords[0].tv = 1.0f;

   Square[1] = CVector4(1.0f, 0.0f,  1.0f);
   SquareCoords[1].tu = 1.0f; SquareCoords[1].tv = 0.0f;

   Square[2] = CVector4(-1.0f, 0.0f,  1.0f);
   SquareCoords[2].tu = 0.0f; SquareCoords[2].tv = 0.0f;

   Square[3] = CVector4(-1.0f, 2.0f, 1.0f);
   SquareCoords[3].tu = 0.0f; SquareCoords[3].tv = 1.0f;


   // First we will test if the user has GL_NV_occulsion_query extension.
   char *extension = (char*)glGetString(GL_EXTENSIONS);

   // If not then we can't go on with the demo.
	if(strstr(extension, "HP_occlusion_test") == NULL) return false;

   // Set occlusion support to true.
   OcclusionSupported = true;

   // Set the initial position of each object.
   ObjectPositions[TORUS]  = CVector4(-2.0f, 0.8f, -0.2f);
   ObjectPositions[CUBE]   = CVector4(0.0f, 0.8f, -0.2f);
   ObjectPositions[SPHERE] = CVector4(2.0f, 0.8f, -0.2f);
   ObjectPositions[SQUARE] = CVector4(0.0f, 1.0f, 1.0f);

   // The sizes for each object.  This is not the size we are sending into the functions that
   // create them because the bounding box will, for some objects, be bigger than the geometry.
   ObjectSizes[CUBE]    = 0.5f;
   ObjectSizes[TORUS]   = 0.6f;
   ObjectSizes[SPHERE]  = 0.7f;
   ObjectSizes[SQUARE]  = 1.0f;

   for(int i = 0; i < MAX_OBJECTS; i++)
      {
         // This will loop through all of the objects and calculate the bounding
         // box based on the objects default position and its size.
         ObjectsBBox[i].min.x = ObjectPositions[i].x - ObjectSizes[i];
         ObjectsBBox[i].min.y = ObjectPositions[i].y - ObjectSizes[i];
         ObjectsBBox[i].min.z = ObjectPositions[i].z - ObjectSizes[i];

         ObjectsBBox[i].max.x = ObjectPositions[i].x + ObjectSizes[i];
         ObjectsBBox[i].max.y = ObjectPositions[i].y + ObjectSizes[i];
         ObjectsBBox[i].max.z = ObjectPositions[i].z + ObjectSizes[i];
      }

   // If all went well we return true.
   return true;
}


void RenderScene()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);   // Clears the screen.
   glLoadIdentity();                   //Reset modelview matrix for new frame.

   // Apply all translations and rotations.
	glTranslatef(0.0f, -2.0f, -10.0f);
	glRotatef(-yRotation, 1.0f, 0.0f, 0.0f);
	glRotatef(-xRotation, 0.0f, 1.0f, 0.0f);

   // Render the floor...
	glPushMatrix();

		DrawGround();
		glBindTexture(GL_TEXTURE_2D, NULL);

	glPopMatrix();
   
   // Loop throught and render all objects.  But first call the RenderTest() function to
   // see if the object is visible (not total behind another object).  If we can see it
   // we render that object and save "True" in the Visible array for that object.
   for(int i = 0; i < MAX_OBJECTS; i++)
      {
         if(RenderTest(ObjectsBBox[i]))
            {
               sprintf(Visible[i], "True");
               Render(i);

               // Render the bounding boxes to the screen if true.  This is for output and
               // has nothing to do with the test anymore.
               if(RenderBoundingBoxes)
                  DrawBoundingBox(ObjectsBBox[i]);
            }
         else
            sprintf(Visible[i], "False");
      }
   

   glDisable(GL_LIGHTING);             // Disable lighting when drawing the text output.
   glBindTexture(GL_TEXTURE_2D, NULL); // Disable the current texture so it doesn't effect text.
   glColor3f(1.0f, 0.0f, 0.0f);        // Set color to red for the first string.

   // The rest of this function deals with printing text output to the screen for our benefit.
   font.PrintText("www.UltimateGameProgramming.com", -0.9f, 0.9f);
   glColor3f(1.0f, 1.0f, 1.0f);        // Change the color back to white.

   font.PrintText("Created by the Programming Ace...", -0.9f, 0.8f);
   font.PrintText("Hp Occlusion demo...", -0.9f, 0.7f);

   font.PrintText("Which are visible...", -0.9f, 0.5f);
   
   sprintf(output, "Cube Visible = %s",Visible[CUBE]);
   font.PrintText(output, -0.9f, 0.4f);
   
   sprintf(output, "Sphere Visible = %s",Visible[SPHERE]);
   font.PrintText(output, -0.9f, 0.3f);
   
   sprintf(output, "Torus Visible = %s",Visible[TORUS]);
   font.PrintText(output, -0.9f, 0.2f);
   
   sprintf(output, "Square Visible = %s",Visible[SQUARE]);
   font.PrintText(output, -0.9f, 0.1f);

   glEnable(GL_LIGHTING);              // enable lighting back.

	SwapBuffers(g_HDC);                 // Display the new frame.
}


void Render(int i)
{
   // Here we render each object.  I used int i to determine which object we want to draw.
   // This makes things easier because we can't draw the whole scene at once because we want
   // to test each object individually if it can be seen or not.

   // Draw the square.
   if(i == SQUARE)
      {
         // Draw the textured square in front of the cube.
         glPushMatrix();

		      glBindTexture(GL_TEXTURE_2D, SquareTex.ID);

		      glBegin(GL_QUADS);

			      glNormal3f(SquareNormal.x, SquareNormal.y, SquareNormal.z);

               for(int i = 0; i < MAX_POINTS; i++)
                  {
                     glTexCoord2f(SquareCoords[i].tu, SquareCoords[i].tv);
                     glVertex3f(Square[i].x, Square[i].y, Square[i].z);
                  }

		      glEnd();

	      glPopMatrix();

         glBindTexture(GL_TEXTURE_2D, NULL);
      }

   // Draw the sphere.
   if(i == SPHERE)
      {
         // Draw the sphere next to the cube.
	      glPushMatrix();

		      glTranslatef(2.0f, 0.8f, -0.2f);
            auxSolidSphere(0.7f);

	      glPopMatrix();
      }
   
   // Draw the torus.
   if(i == TORUS)
      {
	      // From here on out we will draw the rest of the objects starting with the torus.
	      glPushMatrix();

		      glTranslatef(-2.0f, 0.8f, -0.2f);
		      auxSolidTorus(0.2f, 0.4f);

	      glPopMatrix();
      }

   // Last but not least draw the cube.
   if(i == CUBE)
      {
         // Next draw the cube next to the torus.
	      glPushMatrix();

		      glTranslatef(0.0f, 0.8f, -0.2f);
	         auxSolidCube(1.1f);

	      glPopMatrix();
      }
}


bool RenderTest(BoundingBox &Box)
{
   // First we enable the test.  Next we draw the bouding geometry as if we was drawing
   // it to the screen.  Then we disable the test and call
   // glGetBooleanv(GL_OCCLUSION_TEST_RESULT_HP, &result); to get whether or not the bounding
   // box has had any effect of the depth buffer.  If so then we can see it, if not then we can't.

   unsigned char result;                                 // Stores the result of test.

   glDepthMask(GL_FALSE);                                // Optional.
   glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);  // Optional.

   glEnable(GL_OCCLUSION_TEST_HP);                       // Enable the occlusion testing.

   DrawBoundingBox(Box);                                 // Draw bounding box to test.

   glDisable(GL_OCCLUSION_TEST_HP);                      // Disable the occlusion testing.

   glGetBooleanv(GL_OCCLUSION_TEST_RESULT_HP, &result);  // Get the results and store it.

   glDepthMask(GL_TRUE);                                 // Optional (if you didn't set to false).
   glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);      // Optional (if you didn't set to false).

   // If true was returned from the test then we return true to indicate that we can see this
   // object, else false that we can not.
   if(result)
      return true;

   return false;
}


void DrawGround()
{
   // This is the same function from past demos...
   // Bind the terrain texture.
   
	glDisable(GL_LIGHTING);
	glPushMatrix();

	   glBindTexture(GL_TEXTURE_2D, Terrain.ID);

      // Draw one big square.
      glBegin(GL_QUADS);
		   glNormal3f(0.0f, 1.0f, 0.0f);
         glTexCoord2f(0.0, 0.0); glVertex3f(15.0f, -0.01f, 15.0f);
         glTexCoord2f(1.0, 0.0); glVertex3f(-15.0f, -0.01f, 15.0f);
         glTexCoord2f(1.0, 1.0); glVertex3f(-15.0f, -0.01f, -15.0f);
         glTexCoord2f(0.0, 1.0); glVertex3f(15.0f, -0.01f, -15.0f);
      glEnd();

	glPopMatrix();
	glEnable(GL_LIGHTING);
}


// All this function will do is allow us to draw the bounding box.
void DrawBoundingBox(BoundingBox &Box)
{
	glColor3f(1.0, 1.0, 1.0f);
   glBindTexture(GL_TEXTURE_2D, NULL);

   // This will render the bounding box by drawing it as a cube.
   // This code was taken out of the Bouding Box collision detection demo.
	glBegin(GL_LINE_LOOP);
	   glVertex3f(Box.min.x, Box.min.y, Box.min.z);
	   glVertex3f(Box.max.x, Box.min.y, Box.min.z);
	   glVertex3f(Box.max.x, Box.min.y, Box.max.z);
	   glVertex3f(Box.min.x, Box.min.y, Box.max.z);
	glEnd();

	glBegin(GL_LINE_LOOP);
	   glVertex3f(Box.min.x, Box.max.y, Box.min.z);
	   glVertex3f(Box.max.x, Box.max.y, Box.min.z);
	   glVertex3f(Box.max.x, Box.max.y, Box.max.z);
	   glVertex3f(Box.min.x, Box.max.y, Box.max.z);
	glEnd();

	glBegin(GL_LINES);
	   glVertex3f(Box.min.x, Box.min.y, Box.min.z);
	   glVertex3f(Box.min.x, Box.max.y, Box.min.z);
	   glVertex3f(Box.max.x, Box.min.y, Box.min.z);
	   glVertex3f(Box.max.x, Box.max.y, Box.min.z);
   glEnd();

   glBegin(GL_LINES);
	   glVertex3f(Box.max.x, Box.min.y, Box.max.z);
	   glVertex3f(Box.max.x, Box.max.y, Box.max.z);
	   glVertex3f(Box.min.x, Box.min.y, Box.max.z);
	   glVertex3f(Box.min.x, Box.max.y, Box.max.z);
	glEnd();
}


void Shutdown()
{
   // Free both textures that we loaded earlier.
   Terrain.FreeImage();
   SquareTex.FreeImage();

   // Clear the font object and all it used up.
   font.ClearGLFont();
}


// Recap:

// Well as I said before this way of occlusion testing is not nearly as useful
// as the NV_occlusion_query.  But depending on what your doing it might be helpful
// to your needs.  Occlusion using the HP_occlusion_test has less code to it but in this
// demo it seems like there is a lot more only because of the way I set things up and
// rendered the objects.  If you do deciede to use this then I suggest you combine it with
// frustum culling.  That way you can for sure reject a lot of geometry.  The ones that
// are in the frustum you can use this occlusion to determine if it is actually visible.
// Keep in mind that NV_occlusion_query is still the better of the two no matter how I
// keep trying to milk this HP_occlusion_test.


// Copyright October 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com